#include "cache.h"

struct cache {
  unsigned int capacity;
  void **bucket;
  unsigned int (*hashcode)(const void *);
  void *(*data_get)(const void *);
  int (*data_release)(void *);
};

Cache *cache_open(unsigned int capacity, unsigned int (*hashcode)(const void *), void *(*data_get)(const void *), int (*data_release)(void *))
{
  Cache *cache;
  void **bucket;

  cache = XMALLOC(Cache, 1);
  if (cache == NULL) {
    return NULL;
  }
  bucket = XMALLOC(void *, capacity);
  if (bucket == NULL) {
    XFREE(cache);
    return NULL;
  }
  for (int i = 0; i < capacity; i++) {
    bucket[i] = NULL;
  }
  cache->capacity = capacity;
  cache->bucket = bucket;
  cache->hashcode = hashcode;
  cache->data_get = data_get;
  cache->data_release = data_release;
  return cache;
}

int cache_close(Cache *cache)
{
  int rc;

  if (cache == NULL) {
    return FALSE;
  }
  rc = cache_clear(cache);
  XFREE(cache->bucket);
  XFREE(cache);
  return rc;
}

int cache_clear(Cache *cache)
{
  size_t capacity;
  void **bucket;
  int rc;

  rc = TRUE;
  capacity = cache->capacity;
  bucket = cache->bucket;
  for (int i = 0; i < capacity; i++) {
    if (bucket[i] != NULL) {
      if (!cache->data_release(bucket[i])) {
        rc = FALSE;
      }
    }
  }
  return rc;
}

void *cache_tryget(Cache *cache, const void *arg)
{
  unsigned int code;
  void *data;

  code = cache->hashcode(arg)%cache->capacity;
  data = cache->bucket[code];
  return data;
}

void *cache_get(Cache *cache, const void *arg)
{
  unsigned int code;
  void *data;

  code = cache->hashcode(arg)%cache->capacity;
  data = cache->bucket[code];
  if (data == NULL) {
    data = cache->data_get(arg);
    cache->bucket[code] = data;
  }
  return data;
}
